# Benchmarks from https://github.com/deshaw/versioned-hdf5/issues/167

import h5py
import numpy as np

from versioned_hdf5 import VersionedHDF5File

dt = np.dtype("double")


def time_many_chunks():
    d0 = 2
    d1 = 15220
    d2 = 2
    shape = (d0, d1, d2)
    chunks = (600, 2, 4)
    with h5py.File("foo.h5", "w") as f:
        vf = VersionedHDF5File(f)
        with vf.stage_version("0") as sv:
            sv.create_dataset(
                "bar",
                shape=shape,
                maxshape=(None, None, None),
                chunks=chunks,
                dtype=dt,
                data=np.full(shape, 0, dtype=dt),
            )

    i = 1
    with h5py.File("foo.h5", "r+") as f:
        vf = VersionedHDF5File(f)
        with vf.stage_version(str(i)) as sv:
            sv["bar"][:] = np.full(shape, i, dtype=dt)


def time_many_chunks_integer_index():
    d0 = 2
    d1 = 15220
    d2 = 2
    shape = (d0, d1, d2)
    chunks = (600, 2, 4)
    with h5py.File("foo.h5", "w") as f:
        vf = VersionedHDF5File(f)
        with vf.stage_version("0") as sv:
            sv.create_dataset(
                "bar",
                shape=shape,
                maxshape=(None, None, None),
                chunks=chunks,
                dtype=dt,
                data=np.full(shape, 0, dtype=dt),
            )

    i = 1
    with h5py.File("foo.h5", "r+") as f:
        vf = VersionedHDF5File(f)
        with vf.stage_version(str(i)) as sv:
            i2 = np.random.choice(d1, 30, replace=False)
            i2 = np.sort(i2)
            sv["bar"][:, i2, :] = np.full((d0, len(i2), d2), i, dtype=dt)


def time_many_chunks_arange():
    d0 = 2
    d1 = 15220
    d2 = 2
    shape = (d0, d1, d2)
    chunks = (600, 2, 4)
    with h5py.File("foo.h5", "w") as f:
        vf = VersionedHDF5File(f)
        with vf.stage_version("0") as sv:
            sv.create_dataset(
                "bar",
                shape=shape,
                maxshape=(None, None, None),
                chunks=chunks,
                dtype=dt,
                data=np.arange(np.prod(shape), dtype=dt).reshape(shape),
            )



import importlib.metadata
import pathlib
import shutil
import subprocess
from unittest import mock

import h5py
import numpy as np
import pytest
from packaging.version import Version

from versioned_hdf5 import VersionedHDF5File
from versioned_hdf5.hashtable import Hashtable
from versioned_hdf5.replay import (
    _get_parent,
    _recreate_hashtable,
    _recreate_raw_data,
    _recreate_virtual_dataset,
    delete_version,
    delete_versions,
    modify_metadata,
)


def setup_vfile(file):
    with file.stage_version("version1") as g:
        data = g.create_dataset(
            "test_data", data=None, fillvalue=1.0, shape=(10000,), chunks=(1000,)
        )
        data[0] = 0.0
        g.create_dataset("test_data2", data=[1, 2, 3], chunks=(1000,))
        group = g.create_group("group")
        group.create_dataset("test_data4", data=[1, 2, 3, 4], chunks=(1000,))

    with file.stage_version("version2") as g:
        g["test_data"][2000] = 2.0
        g.create_dataset("test_data3", data=[1, 2, 3, 4], chunks=(1000,))
        g["group"]["test_data4"][0] = 5


def check_data(file, test_data_fillvalue=1.0, version2=True, test_data4_fillvalue=0):
    assert set(file["version1"]) == {"test_data", "test_data2", "group"}
    assert file["version1"]["test_data"].shape == (10000,)
    assert file["version1"]["test_data"][0] == 0.0
    assert np.all(file["version1"]["test_data"][1:] == test_data_fillvalue)

    if version2:
        assert set(file["version2"]) == {
            "test_data",
            "test_data2",
            "test_data3",
            "group",
        }
        assert file["version2"]["test_data"].shape == (10000,)
        assert file["version2"]["test_data"][0] == 0.0
        assert np.all(file["version2"]["test_data"][1:2000] == test_data_fillvalue)
        assert file["version2"]["test_data"][2000] == 2.0
        assert np.all(file["version2"]["test_data"][2001:] == test_data_fillvalue)

    assert file["version1"]["test_data2"].shape == (3,)
    assert np.all(file["version1"]["test_data2"][:] == [1, 2, 3])

    if version2:
        assert file["version2"]["test_data2"].shape == (3,)
        assert np.all(file["version2"]["test_data2"][:] == [1, 2, 3])

    assert "test_data3" not in file["version1"]

    if version2:
        assert file["version2"]["test_data3"].shape == (4,)
        assert np.all(file["version2"]["test_data3"][:] == [1, 2, 3, 4])

    assert set(file["version1"]["group"]) == {"test_data4"}
    assert file["version1"]["group"]["test_data4"].shape == (4,)
    np.testing.assert_equal(file["version1"]["group"]["test_data4"][:4], [1, 2, 3, 4])
    assert np.all(file["version1"]["group"]["test_data4"][4:] == test_data4_fillvalue)

    if version2:
        assert set(file["version2"]["group"]) == {"test_data4"}
        assert file["version2"]["group"]["test_data4"].shape == (4,)
        np.testing.assert_equal(
            file["version2"]["group"]["test_data4"][:4], [5, 2, 3, 4]
        )
        assert np.all(
            file["version2"]["group"]["test_data4"][4:] == test_data4_fillvalue
        )


def test_modify_metadata_compression(vfile):
    setup_vfile(vfile)

    f = vfile.f

    assert vfile["version1"]["test_data"].compression == None
    assert vfile["version2"]["test_data"].compression == None
    assert vfile["version1"]["test_data"].compression_opts == None
    assert vfile["version2"]["test_data"].compression_opts == None

    assert vfile["version1"]["test_data2"].compression == None
    assert vfile["version2"]["test_data2"].compression == None
    assert vfile["version1"]["test_data2"].compression_opts == None
    assert vfile["version2"]["test_data2"].compression_opts == None

    assert vfile["version2"]["test_data3"].compression == None
    assert vfile["version2"]["test_data3"].compression_opts == None

    assert vfile["version1"]["group"]["test_data4"].compression == None
    assert vfile["version2"]["group"]["test_data4"].compression == None
    assert vfile["version1"]["group"]["test_data4"].compression_opts == None
    assert vfile["version2"]["group"]["test_data4"].compression_opts == None

    assert f["_version_data"]["test_data"]["raw_data"].compression == None
    assert f["_version_data"]["test_data2"]["raw_data"].compression == None
    assert f["_version_data"]["test_data3"]["raw_data"].compression == None
    assert f["_version_data"]["group"]["test_data4"]["raw_data"].compression == None

    assert f["_version_data"]["test_data"]["raw_data"].compression_opts == None
    assert f["_version_data"]["test_data2"]["raw_data"].compression_opts == None
    assert f["_version_data"]["test_data3"]["raw_data"].compression_opts == None
    assert (
        f["_version_data"]["group"]["test_data4"]["raw_data"].compression_opts == None
    )

    modify_metadata(f, "test_data2", compression="gzip", compression_opts=3)
    check_data(vfile)

    assert vfile["version1"]["test_data"].compression == None
    assert vfile["version2"]["test_data"].compression == None
    assert vfile["version1"]["test_data"].compression_opts == None
    assert vfile["version2"]["test_data"].compression_opts == None

    assert vfile["version1"]["test_data2"].compression == "gzip"
    assert vfile["version2"]["test_data2"].compression == "gzip"
    assert vfile["version1"]["test_data2"].compression_opts == 3
    assert vfile["version2"]["test_data2"].compression_opts == 3

    assert vfile["version2"]["test_data3"].compression == None
    assert vfile["version2"]["test_data3"].compression_opts == None

    assert vfile["version1"]["group"]["test_data4"].compression == None
    assert vfile["version2"]["group"]["test_data4"].compression == None
    assert vfile["version1"]["group"]["test_data4"].compression_opts == None
    assert vfile["version2"]["group"]["test_data4"].compression_opts == None

    assert f["_version_data"]["test_data"]["raw_data"].compression == None
    assert f["_version_data"]["test_data2"]["raw_data"].compression == "gzip"
    assert f["_version_data"]["test_data3"]["raw_data"].compression == None
    assert f["_version_data"]["group"]["test_data4"]["raw_data"].compression == None

    assert f["_version_data"]["test_data"]["raw_data"].compression_opts == None
    assert f["_version_data"]["test_data2"]["raw_data"].compression_opts == 3
    assert f["_version_data"]["test_data3"]["raw_data"].compression_opts == None
    assert (
        f["_version_data"]["group"]["test_data4"]["raw_data"].compression_opts == None
    )

    # Make sure the tmp group group has been destroyed.
    assert set(f["_version_data"]) == {
        "test_data",
        "test_data2",
        "test_data3",
        "group",
        "versions",
    }
    assert set(f["_version_data"]["group"]) == {"test_data4"}


def test_modify_metadata_compression2(vfile):
    setup_vfile(vfile)

    f = vfile.f

    assert vfile["version1"]["test_data"].compression == None
    assert vfile["version2"]["test_data"].compression == None
    assert vfile["version1"]["test_data"].compression_opts == None
    assert vfile["version2"]["test_data"].compression_opts == None

    assert vfile["version1"]["test_data2"].compression == None
    assert vfile["version2"]["test_data2"].compression == None
    assert vfile["version1"]["test_data2"].compression_opts == None
    assert vfile["version2"]["test_data2"].compression_opts == None

    assert vfile["version2"]["test_data3"].compression == None
    assert vfile["version2"]["test_data3"].compression_opts == None

    assert vfile["version1"]["group"]["test_data4"].compression == None
    assert vfile["version2"]["group"]["test_data4"].compression == None
    assert vfile["version1"]["group"]["test_data4"].compression_opts == None
    assert vfile["version2"]["group"]["test_data4"].compression_opts == None

    assert f["_version_data"]["test_data"]["raw_data"].compression == None
    assert f["_version_data"]["test_data2"]["raw_data"].compression == None
    assert f["_version_data"]["test_data3"]["raw_data"].compression == None
    assert f["_version_data"]["group"]["test_data4"]["raw_data"].compression == None

    assert f["_version_data"]["test_data"]["raw_data"].compression_opts == None
    assert f["_version_data"]["test_data2"]["raw_data"].compression_opts == None
    assert f["_version_data"]["test_data3"]["raw_data"].compression_opts == None
    assert (
        f["_version_data"]["group"]["test_data4"]["raw_data"].compression_opts == None
    )

    modify_metadata(f, "group/test_data4", compression="gzip", compression_opts=3)
    check_data(vfile)

    assert vfile["version1"]["test_data"].compression == None
    assert vfile["version2"]["test_data"].compression == None
    assert vfile["version1"]["test_data"].compression_opts == None
    assert vfile["version2"]["test_data"].compression_opts == None

    assert vfile["version1"]["test_data2"].compression == None
    assert vfile["version2"]["test_data2"].compression == None
    assert vfile["version1"]["test_data2"].compression_opts == None
    assert vfile["version2"]["test_data2"].compression_opts == None

    assert vfile["version2"]["test_data3"].compression == None
    assert vfile["version2"]["test_data3"].compression_opts == None

    assert vfile["version1"]["group"]["test_data4"].compression == "gzip"
    assert vfile["version2"]["group"]["test_data4"].compression == "gzip"
    assert vfile["version1"]["group"]["test_data4"].compression_opts == 3
    assert vfile["version2"]["group"]["test_data4"].compression_opts == 3

    assert f["_version_data"]["test_data"]["raw_data"].compression == None
    assert f["_version_data"]["test_data2"]["raw_data"].compression == None
    assert f["_version_data"]["test_data3"]["raw_data"].compression == None
    assert f["_version_data"]["group"]["test_data4"]["raw_data"].compression == "gzip"

    assert f["_version_data"]["test_data"]["raw_data"].compression_opts == None
    assert f["_version_data"]["test_data2"]["raw_data"].compression_opts == None
    assert f["_version_data"]["test_data3"]["raw_data"].compression_opts == None
    assert f["_version_data"]["group"]["test_data4"]["raw_data"].compression_opts == 3

    # Make sure the tmp group group has been destroyed.
    assert set(f["_version_data"]) == {
        "test_data",
        "test_data2",
        "test_data3",
        "group",
        "versions",
    }
    assert set(f["_version_data"]["group"]) == {"test_data4"}


def test_modify_metadata_chunks(vfile):
    setup_vfile(vfile)

    f = vfile.f

    assert vfile["version1"]["test_data"].chunks == (1000,)
    assert vfile["version2"]["test_data"].chunks == (1000,)

    assert vfile["version1"]["test_data2"].chunks == (1000,)
    assert vfile["version2"]["test_data2"].chunks == (1000,)

    assert vfile["version2"]["test_data3"].chunks == (1000,)

    assert vfile["version1"]["group"]["test_data4"].chunks == (1000,)
    assert vfile["version2"]["group"]["test_data4"].chunks == (1000,)

    assert f["_version_data"]["test_data"]["raw_data"].chunks == (1000,)
    assert f["_version_data"]["test_data2"]["raw_data"].chunks == (1000,)
    assert f["_version_data"]["test_data3"]["raw_data"].chunks == (1000,)
    assert f["_version_data"]["group"]["test_data4"]["raw_data"].chunks == (1000,)

    modify_metadata(f, "test_data2", chunks=(500,))
    check_data(vfile)

    assert vfile["version1"]["test_data"].chunks == (1000,)
    assert vfile["version2"]["test_data"].chunks == (1000,)

    assert vfile["version1"]["test_data2"].chunks == (500,)
    assert vfile["version2"]["test_data2"].chunks == (500,)

    assert vfile["version2"]["test_data3"].chunks == (1000,)

    assert vfile["version1"]["group"]["test_data4"].chunks == (1000,)
    assert vfile["version2"]["group"]["test_data4"].chunks == (1000,)

    assert f["_version_data"]["test_data"]["raw_data"].chunks == (1000,)
    assert f["_version_data"]["test_data2"]["raw_data"].chunks == (500,)
    assert f["_version_data"]["test_data3"]["raw_data"].chunks == (1000,)
    assert f["_version_data"]["group"]["test_data4"]["raw_data"].chunks == (1000,)

    # Make sure the tmp group group has been destroyed.
    assert set(f["_version_data"]) == {
        "test_data",
        "test_data2",
        "test_data3",
        "group",
        "versions",
    }
    assert set(f["_version_data"]["group"]) == {"test_data4"}


def test_modify_metadata_chunk2(vfile):
    setup_vfile(vfile)

    f = vfile.f

    assert vfile["version1"]["test_data"].chunks == (1000,)
    assert vfile["version2"]["test_data"].chunks == (1000,)

    assert vfile["version1"]["test_data2"].chunks == (1000,)
    assert vfile["version2"]["test_data2"].chunks == (1000,)

    assert vfile["version2"]["test_data3"].chunks == (1000,)

    assert vfile["version1"]["group"]["test_data4"].chunks == (1000,)
    assert vfile["version2"]["group"]["test_data4"].chunks == (1000,)

    assert f["_version_data"]["test_data"]["raw_data"].chunks == (1000,)
    assert f["_version_data"]["test_data2"]["raw_data"].chunks == (1000,)
    assert f["_version_data"]["test_data3"]["raw_data"].chunks == (1000,)
    assert f["_version_data"]["group"]["test_data4"]["raw_data"].chunks == (1000,)

    modify_metadata(f, "group/test_data4", chunks=(500,))
    check_data(vfile)

    assert vfile["version1"]["test_data"].chunks == (1000,)
    assert vfile["version2"]["test_data"].chunks == (1000,)

    assert vfile["version1"]["test_data2"].chunks == (1000,)
    assert vfile["version2"]["test_data2"].chunks == (1000,)

    assert vfile["version2"]["test_data3"].chunks == (1000,)

    assert vfile["version1"]["group"]["test_data4"].chunks == (500,)
    assert vfile["version2"]["group"]["test_data4"].chunks == (500,)

    assert f["_version_data"]["test_data"]["raw_data"].chunks == (1000,)
    assert f["_version_data"]["test_data2"]["raw_data"].chunks == (1000,)
    assert f["_version_data"]["test_data3"]["raw_data"].chunks == (1000,)
    assert f["_version_data"]["group"]["test_data4"]["raw_data"].chunks == (500,)

    # Make sure the tmp group group has been destroyed.
    assert set(f["_version_data"]) == {
        "test_data",
        "test_data2",
        "test_data3",
        "group",
        "versions",
    }
    assert set(f["_version_data"]["group"]) == {"test_data4"}


def test_modify_metadata_dtype(vfile):
    setup_vfile(vfile)

    f = vfile.f

    assert vfile["version1"]["test_data"].dtype == np.float64
    assert vfile["version2"]["test_data"].dtype == np.float64

    assert vfile["version1"]["test_data2"].dtype == np.int64
    assert vfile["version2"]["test_data2"].dtype == np.int64

    assert vfile["version2"]["test_data3"].dtype == np.int64

    assert vfile["version1"]["group"]["test_data4"].dtype == np.int64
    assert vfile["version2"]["group"]["test_data4"].dtype == np.int64

    assert f["_version_data"]["test_data"]["raw_data"].dtype == np.float64
    assert f["_version_data"]["test_data2"]["raw_data"].dtype == np.int64
    assert f["_version_data"]["test_data3"]["raw_data"].dtype == np.int64
    assert f["_version_data"]["group"]["test_data4"]["raw_data"].dtype == np.int64

    modify_metadata(f, "test_data2", dtype=np.float64)
    check_data(vfile)

    assert vfile["version1"]["test_data"].dtype == np.float64
    assert vfile["version2"]["test_data"].dtype == np.float64

    assert vfile["version1"]["test_data2"].dtype == np.float64
    assert vfile["version2"]["test_data2"].dtype == np.float64

    assert vfile["version2"]["test_data3"].dtype == np.int64

    assert vfile["version1"]["group"]["test_data4"].dtype == np.int64
    assert vfile["version2"]["group"]["test_data4"].dtype == np.int64

    assert f["_version_data"]["test_data"]["raw_data"].dtype == np.float64
    assert f["_version_data"]["test_data2"]["raw_data"].dtype == np.float64
    assert f["_version_data"]["test_data3"]["raw_data"].dtype == np.int64
    assert f["_version_data"]["group"]["test_data4"]["raw_data"].dtype == np.int64

    # Make sure the tmp group group has been destroyed.
    assert set(f["_version_data"]) == {
        "test_data",
        "test_data2",
        "test_data3",
        "group",
        "versions",
    }
    assert set(f["_version_data"]["group"]) == {"test_data4"}


def test_modify_metadata_dtype2(vfile):
    setup_vfile(vfile)

    f = vfile.f

    assert vfile["version1"]["test_data"].dtype == np.float64
    assert vfile["version2"]["test_data"].dtype == np.float64

    assert vfile["version1"]["test_data2"].dtype == np.int64
    assert vfile["version2"]["test_data2"].dtype == np.int64

    assert vfile["version2"]["test_data3"].dtype == np.int64

    assert vfile["version1"]["group"]["test_data4"].dtype == np.int64
    assert vfile["version2"]["group"]["test_data4"].dtype == np.int64

    assert f["_version_data"]["test_data"]["raw_data"].dtype == np.float64
    assert f["_version_data"]["test_data2"]["raw_data"].dtype == np.int64
    assert f["_version_data"]["test_data3"]["raw_data"].dtype == np.int64
    assert f["_version_data"]["group"]["test_data4"]["raw_data"].dtype == np.int64

    modify_metadata(f, "group/test_data4", dtype=np.float64)
    check_data(vfile)

    assert vfile["version1"]["test_data"].dtype == np.float64
    assert vfile["version2"]["test_data"].dtype == np.float64

    assert vfile["version1"]["test_data2"].dtype == np.int64
    assert vfile["version2"]["test_data2"].dtype == np.int64

    assert vfile["version2"]["test_data3"].dtype == np.int64

    assert vfile["version1"]["group"]["test_data4"].dtype == np.float64
    assert vfile["version2"]["group"]["test_data4"].dtype == np.float64

    assert f["_version_data"]["test_data"]["raw_data"].dtype == np.float64
    assert f["_version_data"]["test_data2"]["raw_data"].dtype == np.int64
    assert f["_version_data"]["test_data3"]["raw_data"].dtype == np.int64
    assert f["_version_data"]["group"]["test_data4"]["raw_data"].dtype == np.float64

    # Make sure the tmp group group has been destroyed.
    assert set(f["_version_data"]) == {
        "test_data",
        "test_data2",
        "test_data3",
        "group",
        "versions",
    }
    assert set(f["_version_data"]["group"]) == {"test_data4"}


def test_modify_metadata_fillvalue1(vfile):
    setup_vfile(vfile)

    f = vfile.f

    assert vfile["version1"]["test_data"].fillvalue == 1.0
    assert vfile["version2"]["test_data"].fillvalue == 1.0

    assert vfile["version1"]["test_data2"].fillvalue == 0
    assert vfile["version2"]["test_data2"].fillvalue == 0

    assert vfile["version2"]["test_data3"].fillvalue == 0

    assert vfile["version1"]["group"]["test_data4"].fillvalue == 0
    assert vfile["version2"]["group"]["test_data4"].fillvalue == 0

    assert f["_version_data"]["test_data"]["raw_data"].fillvalue == 1.0
    assert f["_version_data"]["test_data2"]["raw_data"].fillvalue == 0
    assert f["_version_data"]["test_data3"]["raw_data"].fillvalue == 0
    assert f["_version_data"]["group"]["test_data4"]["raw_data"].fillvalue == 0

    modify_metadata(f, "test_data", fillvalue=3.0)
    check_data(vfile, test_data_fillvalue=3.0)

    assert vfile["version1"]["test_data"].fillvalue == 3.0
    assert vfile["version2"]["test_data"].fillvalue == 3.0

    assert vfile["version1"]["test_data2"].fillvalue == 0
    assert vfile["version2"]["test_data2"].fillvalue == 0

    assert vfile["version2"]["test_data3"].fillvalue == 0

    assert vfile["version1"]["group"]["test_data4"].fillvalue == 0
    assert vfile["version2"]["group"]["test_data4"].fillvalue == 0

    assert f["_version_data"]["test_data"]["raw_data"].fillvalue == 3.0
    assert f["_version_data"]["test_data2"]["raw_data"].fillvalue == 0
    assert f["_version_data"]["test_data3"]["raw_data"].fillvalue == 0
    assert f["_version_data"]["group"]["test_data4"]["raw_data"].fillvalue == 0

    # Make sure the tmp group group has been destroyed.
    assert set(f["_version_data"]) == {
        "test_data",
        "test_data2",
        "test_data3",
        "group",
        "versions",
    }
    assert set(f["_version_data"]["group"]) == {"test_data4"}


def test_modify_metadata_fillvalue2(vfile):
    setup_vfile(vfile)

    f = vfile.f

    assert vfile["version1"]["test_data"].fillvalue == 1.0
    assert vfile["version2"]["test_data"].fillvalue == 1.0

    assert vfile["version1"]["test_data2"].fillvalue == 0
    assert vfile["version2"]["test_data2"].fillvalue == 0

    assert vfile["version2"]["test_data3"].fillvalue == 0

    assert vfile["version1"]["group"]["test_data4"].fillvalue == 0
    assert vfile["version2"]["group"]["test_data4"].fillvalue == 0

    assert f["_version_data"]["test_data"]["raw_data"].fillvalue == 1.0
    assert f["_version_data"]["test_data2"]["raw_data"].fillvalue == 0
    assert f["_version_data"]["test_data3"]["raw_data"].fillvalue == 0
    assert f["_version_data"]["group"]["test_data4"]["raw_data"].fillvalue == 0

    modify_metadata(f, "test_data2", fillvalue=3)
    check_data(vfile)

    assert vfile["version1"]["test_data"].fillvalue == 1.0
    assert vfile["version2"]["test_data"].fillvalue == 1.0

    assert vfile["version1"]["test_data2"].fillvalue == 3
    assert vfile["version2"]["test_data2"].fillvalue == 3

    assert vfile["version2"]["test_data3"].fillvalue == 0

    assert vfile["version1"]["group"]["test_data4"].fillvalue == 0
    assert vfile["version2"]["group"]["test_data4"].fillvalue == 0

    assert f["_version_data"]["test_data"]["raw_data"].fillvalue == 1.0
    assert f["_version_data"]["test_data2"]["raw_data"].fillvalue == 3
    assert f["_version_data"]["test_data3"]["raw_data"].fillvalue == 0
    assert f["_version_data"]["group"]["test_data4"]["raw_data"].fillvalue == 0

    # Make sure the tmp group group has been destroyed.
    assert set(f["_version_data"]) == {
        "test_data",
        "test_data2",
        "test_data3",
        "group",
        "versions",
    }
    assert set(f["_version_data"]["group"]) == {"test_data4"}


def test_modify_metadata_fillvalue3(vfile):
    setup_vfile(vfile)

    f = vfile.f

    assert vfile["version1"]["test_data"].fillvalue == 1.0
    assert vfile["version2"]["test_data"].fillvalue == 1.0

    assert vfile["version1"]["test_data2"].fillvalue == 0
    assert vfile["version2"]["test_data2"].fillvalue == 0

    assert vfile["version2"]["test_data3"].fillvalue == 0

    assert vfile["version1"]["group"]["test_data4"].fillvalue == 0
    assert vfile["version2"]["group"]["test_data4"].fillvalue == 0

    assert f["_version_data"]["test_data"]["raw_data"].fillvalue == 1.0
    assert f["_version_data"]["test_data2"]["raw_data"].fillvalue == 0
    assert f["_version_data"]["test_data3"]["raw_data"].fillvalue == 0
    assert f["_version_data"]["group"]["test_data4"]["raw_data"].fillvalue == 0

    modify_metadata(f, "group/test_data4", fillvalue=2)
    check_data(vfile)

    assert vfile["version1"]["test_data"].fillvalue == 1.0
    assert vfile["version2"]["test_data"].fillvalue == 1.0

    assert vfile["version1"]["test_data2"].fillvalue == 0
    assert vfile["version2"]["test_data2"].fillvalue == 0

    assert vfile["version2"]["test_data3"].fillvalue == 0

    assert vfile["version1"]["group"]["test_data4"].fillvalue == 2
    assert vfile["version2"]["group"]["test_data4"].fillvalue == 2

    assert f["_version_data"]["test_data"]["raw_data"].fillvalue == 1.0
    assert f["_version_data"]["test_data2"]["raw_data"].fillvalue == 0
    assert f["_version_data"]["test_data3"]["raw_data"].fillvalue == 0
    assert f["_version_data"]["group"]["test_data4"]["raw_data"].fillvalue == 2

    # Make sure the tmp group group has been destroyed.
    assert set(f["_version_data"]) == {
        "test_data",
        "test_data2",
        "test_data3",
        "group",
        "versions",
    }
    assert set(f["_version_data"]["group"]) == {"test_data4"}


def test_delete_version(vfile):
    setup_vfile(vfile)
    f = vfile.f

    delete_version(f, "version2")
    check_data(vfile, version2=False)
    assert list(vfile) == ["version1"]
    assert set(f["_version_data"]) == {"group", "test_data", "test_data2", "versions"}
    assert set(f["_version_data"]["group"]) == {"test_data4"}
    assert not np.isin(2.0, f["_version_data"]["test_data"]["raw_data"][:])
    assert not np.isin(5, f["_version_data"]["group"]["test_data4"]["raw_data"][:])


def test_delete_versions(vfile):
    setup_vfile(vfile)
    with vfile.stage_version("version3") as g:
        g["test_data"][2000] = 3.0
        g.create_dataset("test_data4", data=[1, 2, 3, 4], chunks=(1000,))
    f = vfile.f

    delete_versions(f, ["version2", "version3"])

    check_data(vfile, version2=False)
    assert list(vfile) == ["version1"]
    assert set(f["_version_data"]) == {"group", "test_data", "test_data2", "versions"}
    assert set(f["_version_data"]["group"]) == {"test_data4"}
    assert not np.isin(2.0, f["_version_data"]["test_data"]["raw_data"][:])
    assert not np.isin(5, f["_version_data"]["group"]["test_data4"]["raw_data"][:])


def test_delete_versions_no_data(vfile):
    with vfile.stage_version("version1") as g:
        g.create_dataset(
            "data",
            maxshape=(None, None),
            chunks=(20, 20),
            shape=(5, 5),
            dtype=np.dtype("int8"),
            fillvalue=0,
        )

    with vfile.stage_version("version2") as g:
        g["data"][0] = 1

    f = vfile.f

    delete_versions(f, ["version2"])
    assert list(vfile) == ["version1"]
    assert list(vfile["version1"]) == ["data"]
    assert vfile["version1"]["data"].shape == (5, 5)
    assert np.all(vfile["version1"]["data"][:] == 0)


def test_delete_versions_no_data2(vfile):
    with vfile.stage_version("version1") as g:
        g.create_dataset(
            "data",
            maxshape=(None, None),
            chunks=(20, 20),
            shape=(5, 5),
            dtype=np.dtype("int8"),
            fillvalue=0,
        )

    with vfile.stage_version("version2") as g:
        g["data"][0] = 1

    f = vfile.f

    delete_versions(f, ["version1"])
    assert list(vfile) == ["version2"]
    assert list(vfile["version2"]) == ["data"]
    assert vfile["version2"]["data"].shape == (5, 5)
    assert np.all(vfile["version2"]["data"][1:] == 0)
    assert np.all(vfile["version2"]["data"][0] == 1)


def test_delete_versions_nested_groups(vfile):
    data = []

    with vfile.stage_version("r0") as sv:
        data_group = sv.create_group("group1/group2")
        data.append(np.arange(500))
        data_group.create_dataset(
            "test_data", maxshape=(None,), chunks=(1000), data=data[0]
        )

    for i in range(1, 11):
        with vfile.stage_version(f"r{i}") as sv:
            data.append(np.random.randint(0, 1000, size=500))
            sv["group1"]["group2"]["test_data"][:] = data[-1]

    assert set(vfile) == {
        "r0",
        "r1",
        "r2",
        "r3",
        "r4",
        "r5",
        "r6",
        "r7",
        "r8",
        "r9",
        "r10",
    }
    for i in range(11):
        assert list(vfile[f"r{i}"]) == ["group1"], i
        assert list(vfile[f"r{i}"]["group1"]) == ["group2"]
        assert list(vfile[f"r{i}"]["group1"]["group2"]) == ["test_data"]
        np.testing.assert_equal(
            vfile[f"r{i}"]["group1"]["group2"]["test_data"][:], data[i]
        )

    delete_versions(vfile, ["r3", "r6"])

    assert set(vfile) == {"r0", "r1", "r2", "r4", "r5", "r7", "r8", "r9", "r10"}
    for i in range(11):
        if i in [3, 6]:
            continue
        assert list(vfile[f"r{i}"]) == ["group1"], i
        assert list(vfile[f"r{i}"]["group1"]) == ["group2"]
        assert list(vfile[f"r{i}"]["group1"]["group2"]) == ["test_data"]
        np.testing.assert_equal(
            vfile[f"r{i}"]["group1"]["group2"]["test_data"][:], data[i]
        )


def test_delete_versions_prev_version(vfile):
    with vfile.stage_version("r0") as g:
        g["foo"] = np.array([1, 2, 3])
    for i in range(1, 11):
        with vfile.stage_version(f"r{i}") as g:
            g["foo"][:] = np.array([1, i, 3])

    delete_versions(vfile, ["r1", "r5", "r8"])
    prev_versions = {
        "__first_version__": None,
        "r0": "__first_version__",
        "r2": "r0",
        "r3": "r2",
        "r4": "r3",
        "r6": "r4",
        "r7": "r6",
        "r9": "r7",
        "r10": "r9",
    }

    for v in vfile:
        assert vfile[v].attrs["prev_version"] == prev_versions[v]


def setup2(vfile):
    with vfile.stage_version("version1") as g:
        g.create_dataset(
            "test_data", data=np.arange(20000).reshape((1000, 20)), chunks=(101, 11)
        )

    with vfile.stage_version("version2") as g:
        g["test_data"][::200] = -g["test_data"][::200]


def test_recreate_raw_data(vfile):
    setup2(vfile)
    raw_data = vfile.f["_version_data/test_data/raw_data"][:]
    assert raw_data.shape == (3030, 11)

    chunks_map = _recreate_raw_data(vfile.f, "test_data", ["version1"])
    new_raw_data = vfile.f["_version_data/test_data/raw_data"][:]

    assert len(chunks_map) == 20

    for old, new in chunks_map.items():
        a = raw_data[old.raw]
        b = new_raw_data[new.raw]
        assert a.shape == b.shape
        np.testing.assert_equal(a, b)


def test_recreate_hashtable(vfile):
    setup2(vfile)
    chunks_map = _recreate_raw_data(vfile.f, "test_data", ["version1"])

    # Recreate a separate, independent version, with the dataset as it would
    # be with version1 deleted.
    with vfile.stage_version("version2_2", prev_version="") as g:
        g.create_dataset(
            "test_data2", data=np.arange(20000).reshape((1000, 20)), chunks=(101, 11)
        )
        g["test_data2"][::200] = -g["test_data2"][::200]

    # orig_hashtable = Hashtable(vfile.f, 'test_data')

    _recreate_hashtable(vfile.f, "test_data", chunks_map, tmp=True)

    new_hash_table = Hashtable(vfile.f, "test_data", hash_table_name="_tmp_hash_table")

    new_hash_table2 = Hashtable(vfile.f, "test_data2")
    d1 = dict(new_hash_table)
    d2 = dict(new_hash_table2)
    assert d1.keys() == d2.keys()

    # The exact slices won't be the same because raw data won't be in the same
    # order
    for h in d1:
        np.testing.assert_equal(
            vfile.f["_version_data/test_data/raw_data"][d1[h].raw],
            vfile.f["_version_data/test_data2/raw_data"][d2[h].raw],
        )


def test_recreate_virtual_dataset(vfile):
    setup2(vfile)
    orig_virtual_dataset = vfile.f["_version_data/versions/version2/test_data"][:]

    chunks_map = _recreate_raw_data(vfile.f, "test_data", ["version1"])

    _recreate_hashtable(vfile.f, "test_data", chunks_map, tmp=False)

    _recreate_virtual_dataset(vfile.f, "test_data", ["version2"], chunks_map, tmp=True)

    new_virtual_dataset = vfile.f["_version_data/versions/version2/_tmp_test_data"][:]

    np.testing.assert_equal(orig_virtual_dataset, new_virtual_dataset)


@pytest.mark.parametrize("execution_number", range(30))
def test_delete_versions2(vfile, execution_number):
    setup2(vfile)
    data = np.arange(20000).reshape((1000, 20))
    data[::200] = -data[::200]

    assert vfile["version2"]["test_data"].shape == data.shape

    delete_versions(vfile, ["version1"])

    assert list(vfile) == ["version2"]

    assert list(vfile["version2"]) == ["test_data"]

    assert vfile["version2"]["test_data"].shape == data.shape

    np.testing.assert_equal(vfile["version2"]["test_data"][:], data)

    assert set(vfile.f["_version_data/test_data/raw_data"][:].flat) == set(data.flat)


def test_delete_versions_variable_length_strings(vfile):
    with vfile.stage_version("r0") as sv:
        data = np.array(["foo"], dtype="O")
        sv.create_dataset("bar", data=data, dtype=h5py.string_dtype(encoding="ascii"))

    for i in range(1, 11):
        with vfile.stage_version("r{}".format(i)) as sv:
            sv["bar"].resize((i + 1,))
            sv["bar"][i] = "foo"

    delete_versions(vfile, ["r2", "r4", "r6"])


def test_delete_versions_fillvalue_only_dataset(vfile):
    with vfile.stage_version("r0") as sv:
        sv.create_dataset(
            "fillvalue_only",
            shape=(6,),
            dtype=np.dtype("int64"),
            data=None,
            maxshape=(None,),
            chunks=(10000,),
            fillvalue=0,
        )
        sv.create_dataset(
            "has_data",
            shape=(6,),
            dtype=np.dtype("int64"),
            data=np.arange(6),
            maxshape=(None,),
            chunks=(10000,),
            fillvalue=0,
        )

    with vfile.stage_version("r1") as sv:
        sv["has_data"] = np.arange(5, -1, -1)

    delete_versions(vfile, ["r0"])

    with vfile.stage_version("r2") as sv:
        sv["fillvalue_only"][0] = 1

    assert set(vfile) == {"r1", "r2"}
    assert set(vfile["r1"]) == {"fillvalue_only", "has_data"}
    assert set(vfile["r2"]) == {"fillvalue_only", "has_data"}
    np.testing.assert_equal(vfile["r1"]["fillvalue_only"][:], 0)
    np.testing.assert_equal(
        vfile["r2"]["fillvalue_only"][:], np.array([1, 0, 0, 0, 0, 0])
    )
    np.testing.assert_equal(vfile["r1"]["has_data"][:], np.arange(5, -1, -1))
    np.testing.assert_equal(vfile["r2"]["has_data"][:], np.arange(5, -1, -1))


def test_delete_versions_current_version(vfile):
    with vfile.stage_version("r0") as sv:
        sv.create_dataset("bar", data=np.arange(10))

    for i in range(1, 11):
        with vfile.stage_version("r{}".format(i)) as sv:
            sv["bar"] = np.arange(10 + i)

    delete_versions(vfile, ["r2", "r4", "r6", "r8", "r9", "r10"])

    cv = vfile.current_version
    assert cv == "r7"
    np.testing.assert_equal(vfile[cv]["bar"][:], np.arange(17))


def test_variable_length_strings(vfile):
    # Warning: this test will segfault with h5py 3.7.0
    # (https://github.com/h5py/h5py/pull/2111 fixes it)
    with vfile.stage_version("r0") as sv:
        g = sv.create_group("data")
        dt = h5py.string_dtype(encoding="ascii")
        g.create_dataset("foo", data=["foo", "bar"], dtype=dt)

    for i in range(1, 7):
        with vfile.stage_version(f"r{i}") as sv:
            sv["data/foo"] = np.array([f"foo{i}", f"bar{i}"], dtype="O")

    delete_versions(vfile, ["r1"])


def test_delete_empty_dataset(vfile):
    """Test that deleting an empty dataset executes successfully."""
    with vfile.stage_version("r0") as sv:
        sv.create_dataset(
            "key0",
            data=np.array([]),
            maxshape=(None,),
            chunks=(10000,),
            compression="lzf",
        )

    # Raw data should be filled with fillvalue, but actual current
    # version dataset should have size 0.
    assert vfile.f["_version_data/key0/raw_data"][:].size == 10000
    assert vfile[vfile.current_version]["key0"][:].size == 0

    # Create a new version, checking again the size
    with vfile.stage_version("r1") as sv:
        sv["key0"].resize((0,))
    assert vfile.f["_version_data/key0/raw_data"][:].size == 10000
    assert vfile[vfile.current_version]["key0"][:].size == 0

    # Deleting a prior version should not change the data in the current version
    delete_versions(vfile, ["r0"])
    assert vfile.f["_version_data/key0/raw_data"][:].size == 10000
    assert vfile[vfile.current_version]["key0"][:].size == 0

    # Create a new version, then check if the data is the correct size
    with vfile.stage_version("r2") as sv:
        sv["key0"].resize((0,))

    assert vfile.f["_version_data/key0/raw_data"][:].size == 10000
    assert vfile[vfile.current_version]["key0"][:].size == 0


@pytest.mark.skipif(shutil.which("h5repack") is None, reason="Requires h5repack to run")
def test_delete_string_dataset(filepath):
    """Test that delete_versions + h5repack works correctly for variable length string
    dtypes.

    When calling delete_versions, the dataset must be reconstructed from the remaining
    versions using a NoneType fillvalue. However, because we can't store a NoneType for
    the fillvalue of the dataset in the h5 file, it is instead stored as b''. Previously
    a bug in delete_versions would recreate the datset using the file's fillvalue of
    b'' rather than None, corrupting the data. See https://github.com/h5py/h5py/issues/941
    for more information about the bug in h5py responsible for this, and
    https://github.com/deshaw/versioned-hdf5/issues/238 for the versioned-hdf5
    discussion.
    """
    # Create two versions of variable-length string data, then delete the first,
    # forcing a reconstruction of the data
    with h5py.File(filepath, "w") as f:
        vf = VersionedHDF5File(f)
        with vf.stage_version("r0") as sv:
            sv.create_dataset(
                "foo",
                data=["abc"],
                chunks=(100,),
                dtype=h5py.string_dtype(encoding="ascii"),
            )

        with vf.stage_version("r1") as sv:
            sv["foo"][0] = "def"

        delete_versions(f, "r0")

    # Repack the data; the RuntimeError only appears if you do this
    tmp_path = pathlib.Path(filepath + ".tmp")
    subprocess.run(["h5repack", filepath, tmp_path], stdout=subprocess.PIPE)

    # Check that staging a new version after delete_versions + h5repack works
    with h5py.File(tmp_path, "r+") as f:
        vf = VersionedHDF5File(f)
        with vf.stage_version("r3") as sv:
            pass

    # Delete the tmp file to avoid leaving around stale file
    tmp_path.unlink(missing_ok=True)


def test_delete_versions_speed(vfile):
    """Test that delete_versions only needs linear time to find the previous version
    for the versions that are being kept.
    """
    with vfile.stage_version("r0") as sv:
        sv.create_dataset(
            "values",
            data=np.zeros(100),
            fillvalue=0,
            chunks=(300,),
            maxshape=(None,),
            compression="lzf",
        )

    for i in range(1, 100):
        with vfile.stage_version(f"r{i}") as sv:
            sv["values"][:] = np.arange(i, i + 100)

    # keep only every 10th version
    versions_to_delete = []
    versions = sorted(
        [(v, vfile._versions[v].attrs["timestamp"]) for v in vfile._versions],
        key=lambda t: t[1],
    )
    for i, v in enumerate(versions):
        if i % 10 != 0:
            versions_to_delete.append(v[0])

    with mock.patch(
        "versioned_hdf5.replay._get_parent", wraps=_get_parent
    ) as mock_get_parent:
        delete_versions(vfile, versions_to_delete)

    # There are 90 versions to delete, and 10 to keep. Each of the 10 we are
    # keeping has to go up 9 versions from it's current previous version, for
    # a total of 90 calls.
    assert mock_get_parent.call_count == 90


@pytest.mark.parametrize(
    ("obj", "metadata_opts"),
    [
        ("test_data2", {"compression": "gzip", "compression_opts": 3}),
        ("group/test_data4", {"compression": "gzip", "compression_opts": 3}),
    ],
)
def test_modify_metadata_compression_default_compression(vfile, obj, metadata_opts):
    """Test that setting compression via modify_metadata works for default compression."""
    setup_vfile(vfile)

    f = vfile.f

    # Check that the compression is unset for every dataset
    for dataset in ["test_data", "test_data2", "group/test_data4"]:
        for version in ["version1", "version2"]:
            assert vfile[version][dataset].compression is None
            assert vfile[version][dataset].compression_opts is None

        assert f["_version_data"][dataset]["raw_data"].compression is None
        assert f["_version_data"][dataset]["raw_data"].compression_opts is None

    modify_metadata(f, obj, **metadata_opts)
    check_data(vfile)

    # Check that the compression is set for the group that had its metadata modified
    for dataset in ["test_data", "test_data2", "group/test_data4"]:
        for version in ["version1", "version2"]:
            if dataset == obj:
                assert (
                    vfile[version][dataset].compression == metadata_opts["compression"]
                )
                assert (
                    vfile[version][dataset].compression_opts
                    == metadata_opts["compression_opts"]
                )
            else:
                assert vfile[version][dataset].compression is None
                assert vfile[version][dataset].compression_opts is None

        if dataset == obj:
            assert (
                f["_version_data"][dataset]["raw_data"].compression
                == metadata_opts["compression"]
            )
            assert (
                f["_version_data"][dataset]["raw_data"].compression_opts
                == metadata_opts["compression_opts"]
            )
        else:
            assert f["_version_data"][dataset]["raw_data"].compression is None
            assert f["_version_data"][dataset]["raw_data"].compression_opts is None

    # Make sure the tmp group group has been destroyed.
    assert set(f["_version_data"]) == {
        "test_data",
        "test_data2",
        "test_data3",
        "group",
        "versions",
    }
    assert set(f["_version_data"]["group"]) == {"test_data4"}


@pytest.mark.parametrize(
    ("obj", "metadata_opts"),
    [
        (
            "test_data2",
            {"compression": 32001, "compression_opts": (0, 0, 0, 0, 7, 1, 2)},
        ),
        (
            "group/test_data4",
            {"compression": 32001, "compression_opts": (0, 0, 0, 0, 7, 1, 2)},
        ),
    ],
)
@pytest.mark.parametrize(
    ("library"),
    [
        "hdf5plugin",
        "tables",
    ],
)
def test_modify_metadata_compression_nondefault_compression(
    vfile, obj, metadata_opts, library
):
    """Test that setting compression via modify_metadata works for nondefault compression."""
    if library == "tables" and Version(importlib.metadata.version("numpy")) >= Version(
        "2"
    ):
        pytest.skip("Skipping test; pytables is incompatible with numpy>=2")
    pytest.importorskip(library)

    setup_vfile(vfile)

    f = vfile.f

    # Check that the compression is unset for every dataset
    for dataset in ["test_data", "test_data2", "group/test_data4"]:
        for version in ["version1", "version2"]:
            assert vfile[version][dataset].compression is None
            assert vfile[version][dataset].compression_opts is None

        raw_data = f["_version_data"][dataset]["raw_data"]
        assert raw_data.compression is None
        assert raw_data.compression_opts is None

    modify_metadata(f, obj, **metadata_opts)
    check_data(vfile)

    # Check that the compression is not set for the group that had its metadata modified;
    # the compression of a virtual dataset does not get set from its parent
    for dataset in ["test_data", "test_data2", "group/test_data4"]:
        for version in ["version1", "version2"]:
            if dataset == obj:
                assert vfile[version][dataset].compression is None
                assert vfile[version][dataset].compression_opts is None
            else:
                assert vfile[version][dataset].compression is None
                assert vfile[version][dataset].compression_opts is None

        raw_data = f["_version_data"][dataset]["raw_data"]
        if dataset == obj:
            assert raw_data.compression is None
            assert raw_data.compression_opts is None

            # Ignore the first four values; for blosc (id 32001) they are reserved
            assert (
                raw_data._filters[str(metadata_opts["compression"])][4:]
                == metadata_opts["compression_opts"][4:]
            )
        else:
            assert raw_data.compression is None
            assert raw_data.compression_opts is None

    # Make sure the tmp group group has been destroyed.
    assert set(f["_version_data"]) == {
        "test_data",
        "test_data2",
        "test_data3",
        "group",
        "versions",
    }
    assert set(f["_version_data"]["group"]) == {"test_data4"}



import json
import os

import h5py
from generate_data import TestVersionedDatasetPerformance as TVDP

from versioned_hdf5 import VersionedHDF5File


# auxiliary code to format file sizes
def format_size(size):
    """
    Auxiliary function to convert bytes to a more readable
    human format.
    """
    suffixes = ["B", "KB", "MB", "GB"]
    i = 0
    while size >= 1024 and i < len(suffixes) - 1:
        size = size / 1024
        i += 1
    return f"{size:.2f} {suffixes[i]}"


class PerformanceTests:
    def __init__(self, **kwargs):
        pass

    def _setoptions(self, options):
        keys = options.keys()
        if "path" in keys:
            self.path = options["path"]
        else:
            self.path = "."
        if "num_transactions" in keys:
            self.num_transactions = options["num_transactions"]
        else:
            self.num_transactions = []
        if "exponents" in keys:
            self.exponents = options["exponents"]
        else:
            self.exponents = []
        if "compression" in keys:
            self.compression = options["compression"]
        else:
            self.compression = []
        if "verbose" in keys:
            self.verbose = options["verbose"]
        else:
            self.verbose = False

    def create_files(self, versions=True):
        tests = []
        msg = ""
        for c in self.compression:
            for p in self.exponents:
                for n in self.num_transactions:
                    chunk_size = 2**p
                    if versions:
                        name = f"{self.testname}_{n}_{p}_{c}"
                    else:
                        name = f"{self.testname}_{n}_{p}_{c}_no_versions"
                    filename = os.path.join(self.path, f"{name}.h5")
                    msg += (
                        f"File with {n} transactions, chunk size 2**{p} "
                        f"and compression filter {c}"
                    )
                    try:
                        h5pyfile = h5py.File(filename, "r")
                        msg += " exists - unable to compute creation time.\n"
                        t = 0
                    except Exception:
                        msg += " not available. Creating new file.\n"
                        # t0 = time.time()
                        t = self.testfun(
                            n,
                            name,
                            chunk_size,
                            c,
                            versions=versions,
                            deterministic=True,
                        )
                        # t = time.time()-t0
                        h5pyfile = h5py.File(filename, "r")
                    if versions:
                        data = VersionedHDF5File(h5pyfile)
                        tests.append(
                            dict(
                                num_transactions=n,
                                chunk_size=chunk_size,
                                compression=c,
                                filename=filename,
                                h5pyfile=h5pyfile,
                                data=data,
                                t_write=t,
                            )
                        )
                    else:
                        tests.append(
                            dict(
                                num_transactions=n,
                                chunk_size=chunk_size,
                                compression=c,
                                filename=filename,
                                h5pyfile=h5pyfile,
                                t_write=t,
                            )
                        )

        for test in tests:
            test["size"] = os.path.getsize(test["filename"])
            test["size_label"] = format_size(test["size"])

        if versions:
            nt = len(self.num_transactions)
            for test in tests[-nt:]:
                lengths = []
                total_size = 0
                for vname in test["data"]._versions:
                    if vname != "__first_version__":
                        version = test["data"][vname]
                        group_key = list(version.keys())[0]
                        lengths.append(len(version[group_key]["val"]))
                        total_size += len(version[group_key]["val"])
                test["theoretical_sizes"] = 24 * total_size

        # Removing some irrelevant info from the dictionary
        summary = []
        for test in tests:
            summary.append(
                dict(
                    (k, test[k])
                    for k in [
                        "num_transactions",
                        "filename",
                        "size",
                        "size_label",
                        "t_write",
                        "chunk_size",
                        "compression",
                    ]
                )
            )
            test["h5pyfile"].close()

        self.tests = tests
        return summary, msg

    def save(self, summary, filename):
        with open(f"{filename}.json", "w") as json_out:
            json.dump(summary, json_out)


class test_large_fraction_changes_sparse(PerformanceTests):
    def __init__(self, **kwargs):
        self.testname = "test_large_fraction_changes_sparse"
        self.testfun = TVDP().test_large_fraction_changes_sparse
        super()._setoptions(options=kwargs)

    def create_files(self, versions=True):
        return super().create_files(versions=versions)

    def save(self, summary, filename):
        super().save(summary, filename)


class test_small_fraction_changes_sparse(PerformanceTests):
    def __init__(self, **kwargs):
        self.testname = "test_small_fraction_changes_sparse"
        self.testfun = TVDP().test_small_fraction_changes_sparse
        super()._setoptions(options=kwargs)

    def create_files(self, versions=True):
        return super().create_files(versions=versions)

    def save(self, summary, filename):
        super().save(summary, filename)


class test_mostly_appends_sparse(PerformanceTests):
    def __init__(self, **kwargs):
        self.testname = "test_mostly_appends_sparse"
        self.testfun = TVDP().test_mostly_appends_sparse
        super()._setoptions(options=kwargs)

    def create_files(self, versions=True):
        return super().create_files(versions=versions)

    def save(self, summary, filename):
        super().save(summary, filename)


class test_mostly_appends_dense(PerformanceTests):
    def __init__(self, **kwargs):
        self.testname = "test_mostly_appends_dense"
        self.testfun = TVDP().test_mostly_appends_dense
        super()._setoptions(options=kwargs)

    def create_files(self, versions=True):
        return super().create_files(versions=versions)

    def save(self, summary, filename):
        super().save(summary, filename)


class test_large_fraction_constant_sparse(PerformanceTests):
    def __init__(self, **kwargs):
        self.testname = "test_large_fraction_constant_sparse"
        self.testfun = TVDP().test_large_fraction_constant_sparse
        super()._setoptions(options=kwargs)

    def create_files(self, versions=True):
        return super().create_files(versions=versions)

    def save(self, summary, filename):
        super().save(summary, filename)


if __name__ == "__main__":
    tests = [
        test_mostly_appends_dense,
        test_small_fraction_changes_sparse,
        test_large_fraction_changes_sparse,
        test_large_fraction_constant_sparse,
        test_mostly_appends_sparse,
    ]

    for test in tests:
        testcase = test(
            num_transactions=[2],
            exponents=[
                12,
            ],
            compression=[
                None,
            ],
        )
        summary, msg = testcase.create_files(versions=True)
        testcase.save(summary, f"{testcase.testname}")
        summary, msg = testcase.create_files(versions=False)
        testcase.save(summary, f"{testcase.testname}_no_versions")



# This file allows cimport'ing the functions and types declared below from other
# Cython modules

from libc.stdint cimport uint64_t

# Centralized definition of hsize_t, in accordance with libhd5:
# https://github.com/HDFGroup/hdf5/blob/6b43197b0817596f47670c6b55d26ff7f86d6bd9/src/H5public.h#L301
#
# versioned_hdf5 uses the same datatype for indexing as libhdf5. Notably, this differs
# from numpy's npy_intp (same as Py_ssize_t, ssize_t, and signed long), which is only 32
# bit wide on 32 bit platforms, to allow indexing datasets with >=2**31 points per axis
# on disk, as long as you don't load them in memory all at once.
#
# Note that hsize_t is unsigned, which can lead to integer underflows.
#
# The definition of hsize_t has changed over time in the libhdf5 headers, as noted
# below.
# C99 dictates that long long is *at least* 64 bit wide, while uint64_t is *exactly* 64
# bit wide. So the two definitions are de facto equivalent.
# The below definition is inconsequential, as it is overridden by whatever version of
# hdf5.h is installed thanks to the 'cdef extern from "hdf5.h"' block.
#
# h5py gets this wildly wrong, as it defines hsize_t as long long, which is signed, and
# confusingly also defines hssize_t as signed long long - which is an alias for
# long long:
# https://github.com/h5py/h5py/blob/eaa9d93cc7620f3e7d8cff6f3a739b7d9834aef7/h5py/api_types_hdf5.pxd#L21-L22

cdef extern from "hdf5.h":
    # ctypedef unsigned long long hsize_t  # hdf5 <=1.12
    ctypedef uint64_t hsize_t  # hdf5 >=1.14

cpdef hsize_t stop2count(hsize_t start, hsize_t stop, hsize_t step) noexcept nogil
cpdef hsize_t count2stop(hsize_t start, hsize_t count, hsize_t step) noexcept nogil
cpdef hsize_t ceil_a_over_b(hsize_t a, hsize_t b) noexcept nogil
cpdef hsize_t smallest_step_after(hsize_t x, hsize_t a, hsize_t m) noexcept nogil



from __future__ import annotations

import gc
import logging
import posixpath
import sys
from collections.abc import Iterable
from copy import deepcopy
from typing import Any

import numpy as np
from h5py import Dataset, File, Group, HLObject, VirtualLayout
from h5py import __version__ as h5py_version
from h5py import h5s
from h5py._hl.selections import select
from h5py._hl.vds import VDSmap
from h5py.h5i import get_name
from ndindex import ChunkSize, Slice, Tuple
from ndindex.ndindex import NDIndex

from .api import VersionedHDF5File
from .backend import (
    create_base_dataset,
    create_virtual_dataset,
    initialize,
    write_dataset,
    write_dataset_chunks,
)
from .hashtable import Hashtable
from .slicetools import spaceid_to_slice
from .versions import all_versions
from .wrappers import (
    DatasetWrapper,
    InMemoryArrayDataset,
    InMemoryDataset,
    InMemoryGroup,
    InMemorySparseDataset,
    _groups,
)

logger = logging.getLogger(__name__)

if "sphinx-build" in sys.argv[0]:

    class File:  # type: ignore[no-redef]
        """Stub vs. incomplete h5py documentation"""


def recreate_dataset(f, name, newf, callback=None):
    """
    Recreate dataset from all versions into `newf`

    `newf` should be a versioned hdf5 file/group that is already initialized
    (it may or may not be in the same physical file as f). Typically `newf`
    should be `tmp_group(f)` (see :func:`tmp_group`).

    `callback` should be a function with the signature

        callback(dataset, version_name)

    It will be called on every dataset in every version. It should return the
    dataset to be used for the new version. The dataset and its containing
    group should not be modified in-place. If a new copy of a dataset is to be
    used, it should be one of the dataset classes in versioned_hdf5.wrappers,
    and should placed in a temporary group, which you may delete after
    `recreate_dataset()` is done. The callback may also return None, in which
    case the dataset is deleted for the given version.

    Note: this function is only for advanced usage. Typical use-cases should
    use :func:`delete_version()` or :func:`modify_metadata()`.
    """
    if isinstance(f, VersionedHDF5File):
        f = f.f

    raw_data = f["_version_data"][name]["raw_data"]

    dtype = raw_data.dtype
    chunks = raw_data.chunks
    compression = raw_data.compression
    compression_opts = raw_data.compression_opts
    fillvalue = raw_data.fillvalue

    first = True
    for version_name in all_versions(f):
        if name in f["_version_data/versions"][version_name]:
            group = InMemoryGroup(
                f["_version_data/versions"][version_name].id, _committed=True
            )

            dataset = group[name]
            if callback:
                dataset = callback(dataset, version_name)
                if dataset is None:
                    continue

            dtype = dataset.dtype
            shape = dataset.shape
            chunks = dataset.chunks
            compression = dataset.compression
            compression_opts = dataset.compression_opts

            if compression is None and getattr(dataset, "_filters", None):
                # If we're using nondefault compression, there's no way of knowing
                # whether the first filter is a valid compression or some other
                # kind of filter, so we issue a warning about assuming that it is
                # the dataset's compression.
                compression = list(dataset._filters)[0]
                compression_opts = dataset._filters[compression]
                logger.warning(
                    "No default compression detected in this dataset. "
                    f"Using first filter {compression} and options "
                    f"{compression_opts} for compression."
                )

            fillvalue = dataset.fillvalue
            attrs = dataset.attrs
            if first:
                create_base_dataset(
                    newf,
                    name,
                    data=np.empty((0,) * len(dataset.shape), dtype=dtype),
                    dtype=dtype,
                    chunks=chunks,
                    compression=compression,
                    compression_opts=compression_opts,
                    fillvalue=fillvalue,
                )
                first = False
            # Read in all the chunks of the dataset (we can't assume the new
            # hash table has the raw data in the same locations, even if the
            # data is unchanged).
            if isinstance(dataset, (InMemoryDataset, InMemorySparseDataset)):
                for c, index in dataset.data_dict.copy().items():
                    if isinstance(index, Slice):
                        dataset[c.raw]
                        assert not isinstance(dataset.data_dict[c], Slice)
                slices = write_dataset_chunks(newf, name, dataset.data_dict)
            else:
                slices = write_dataset(newf, name, dataset)
            create_virtual_dataset(
                newf,
                version_name,
                name,
                shape,
                slices,
                attrs=attrs,
                fillvalue=fillvalue,
            )


def tmp_group(f):
    """
    Create a temporary group in `f` for use with :func:`recreate_dataset`.
    """
    if isinstance(f, VersionedHDF5File):
        f = f.f

    if "__tmp__" not in f["_version_data"]:
        tmp = f["_version_data"].create_group("__tmp__")
        initialize(tmp)
        for version_name in all_versions(f):
            group = f["_version_data/versions"][version_name]
            new_group = tmp["_version_data/versions"].create_group(version_name)
            for k, v in group.attrs.items():
                new_group.attrs[k] = v
    else:
        tmp = f["_version_data/__tmp__"]
    return tmp


# See InMemoryDataset.fillvalue. In h5py3 variable length strings use None
# for the h5py fillvalue, but require a string fillvalue for NumPy.
def _get_np_fillvalue(data: Dataset) -> Any:
    """Get the fillvalue for an empty dataset.

    See InMemoryDataset.fillvalue. In h5py3 variable length strings use None
    for the h5py fillvalue, but require a string fillvalue for NumPy.

    Parameters
    ----------
    data : Dataset
        Data for which the fillvalue is to be retrieved

    Returns
    -------
    Any
        Value used to fill the empty dataset; can be any numpy scalar type supported by
        h5py
    """
    if data.fillvalue is not None:
        return data.fillvalue
    if data.dtype.metadata:
        if "vlen" in data.dtype.metadata:
            if h5py_version.startswith("3") and data.dtype.metadata["vlen"] == str:
                return bytes()
            return data.dtype.metadata["vlen"]()
        elif "h5py_encoding" in data.dtype.metadata:
            return data.dtype.type()
    return np.zeros((), dtype=data.dtype)[()]


def _recreate_raw_data(
    f: VersionedHDF5File | File,
    name: str,
    versions_to_delete: Iterable[str],
) -> dict[NDIndex, NDIndex] | None:
    """Create a new raw dataset without the chunks from versions_to_delete.

    Parameters
    ----------
    f : VersionedHDF5File | File
        File for which the raw data is to be reconstructed
    name : str
        Name of the dataset
    versions_to_delete : Iterable[str]
        Versions to omit from the reconstructed raw_data

    Returns
    -------
    dict[NDIndex, NDIndex] | None
        A mapping between old raw dataset chunks and the new raw dataset chunks

        If no chunks would be left, i.e., the dataset does not appear in any
        version not in versions_to_delete, None is returned.
    """
    chunks_to_keep = set()

    if isinstance(f, VersionedHDF5File):
        vf = f
    else:
        vf = VersionedHDF5File(f)

    for version in vf.versions:
        if version not in versions_to_delete and name in vf[version]:
            dataset = f["_version_data/versions"][version][name]

            if dataset.is_virtual:
                for i in dataset.virtual_sources():
                    chunks_to_keep.add(spaceid_to_slice(i.src_space))

    raw_data = f["_version_data"][name]["raw_data"]
    chunks = ChunkSize(raw_data.chunks)
    new_shape = (len(chunks_to_keep) * chunks[0], *chunks[1:])

    fillvalue = _get_np_fillvalue(raw_data)
    # Guard against existing _tmp_raw_data
    _delete_tmp_raw_data(f, name)

    chunks_to_keep = sorted(chunks_to_keep, key=lambda i: i.args[0].args[0])

    raw_data_chunks_map = {}
    for new_chunk, chunk in zip(chunks.indices(new_shape), chunks_to_keep):
        # Truncate the new slice if it isn't a full chunk
        to_set_fillvalue = []
        new_truncated = []
        for i in range(len(new_chunk.args)):
            end = new_chunk.args[i].start + len(chunk.args[i])
            new_truncated.append(Slice(new_chunk.args[i].start, end))

            # If one dimension is truncated, create slices into
            # all other dimensions to be set to the fillvalue
            if len(new_chunk.args[i]) != len(chunk.args[i]):
                to_fill = []
                for j in range(len(new_chunk.args)):
                    if j == i:
                        to_fill.append(Slice(end, new_chunk.args[i].stop))
                    else:
                        to_fill.append(
                            Slice(
                                new_chunk.args[j].start,
                                new_chunk.args[j].start + len(chunk.args[j]),
                            )
                        )
                to_set_fillvalue.append(Tuple(*to_fill))

        new_truncated = Tuple(*new_truncated)
        raw_data[new_truncated.raw] = raw_data[chunk.raw]

        # Set the fillvalue of any slices which were truncated.
        for tup in to_set_fillvalue:
            raw_data[tup.raw] = fillvalue
        raw_data_chunks_map[chunk] = new_truncated

    raw_data.resize(new_shape)

    return raw_data_chunks_map


def _delete_tmp_raw_data(f: File, name: str):
    """Delete _tmp_raw_data if it exists in the file.

    Parameters
    ----------
    f : File
        File in which _tmp_raw_data is to be removed
    name : str
        Name of the dataset where _tmp_raw_data is to be removed
    """
    if "_tmp_raw_data" in f["_version_data"][name]:
        del f["_version_data"][name]["_tmp_raw_data"]


def _recreate_hashtable(f, name, raw_data_chunks_map, tmp=False):
    """
    Recreate the hashtable for the dataset f, with only the new chunks in the
    raw_data_chunks_map.

    If tmp=True, a new hashtable called '_tmp_hash_table' is created.
    Otherwise the hashtable is replaced.
    """
    # We could just reconstruct the hashtable with from_raw_data, but that is
    # slow, so instead we recreate it manually from the old hashable and the
    # raw_data_chunks_map.
    new_hash_table = Hashtable(f, name, hash_table_name="_tmp_hash_table")
    old_inverse = Hashtable(f, name).inverse()

    for old_chunk, new_chunk in raw_data_chunks_map.items():
        if isinstance(old_chunk, Tuple):
            old_chunk = old_chunk.args[0]
        if isinstance(new_chunk, Tuple):
            new_chunk = new_chunk.args[0]

        new_hash_table[old_inverse[old_chunk.reduce()]] = new_chunk

    new_hash_table.write()

    if not tmp:
        del f["_version_data"][name]["hash_table"]
        f["_version_data"][name].move("_tmp_hash_table", "hash_table")


def _recreate_virtual_dataset(f, name, versions, raw_data_chunks_map, tmp=False):
    """
    Recreate every virtual dataset `name` in the versions `versions` according
    to the new raw_data chunks in `raw_data_chunks_map`.

    Returns a dict mapping the chunks from the old raw dataset to the chunks
    in the new raw dataset. Chunks not in the mapping were deleted. If the
    dict is empty, then no remaining version contains the given dataset.

    If tmp is True, the new virtual datasets are named `'_tmp_' + name` and
    are placed alongside the existing ones. Otherwise the existing virtual
    datasets are replaced.

    """
    raw_data = f["_version_data"][name]["raw_data"]

    for version_name in versions:
        if name not in f["_version_data/versions"][version_name]:
            continue

        group = f["_version_data/versions"][version_name]
        dataset = group[name]

        # See the comments in create_virtual_dataset
        layout = VirtualLayout(dataset.shape, dtype=dataset.dtype)
        layout_has_sources = hasattr(layout, "sources")

        if not layout_has_sources:
            from h5py import _selector

            layout._src_filenames.add(b".")
            space = h5s.create_simple(dataset.shape)
            selector = _selector.Selector(space)

        # If a dataset has no data except for the fillvalue, it will not be virtual
        if dataset.is_virtual:
            virtual_sources = dataset.virtual_sources()
            for vmap in virtual_sources:
                vspace, fname, dset_name, src_space = vmap
                fname = fname.encode("utf-8")
                assert fname == b".", fname

                vslice = spaceid_to_slice(vspace)
                src_slice = spaceid_to_slice(src_space)
                if src_slice not in raw_data_chunks_map:
                    raise ValueError(
                        f"Could not find the chunk for {vslice} ({src_slice} in the old raw dataset) for {name!r} in {version_name!r}"
                    )
                new_src_slice = raw_data_chunks_map[src_slice]

                if not layout_has_sources:
                    key = new_src_slice.raw
                    vs_sel = select(raw_data.shape, key, dataset=None)

                    sel = selector.make_selection(vslice.raw)
                    layout.dcpl.set_virtual(
                        sel.id, b".", raw_data.name.encode("utf-8"), vs_sel.id
                    )
                else:
                    vs_sel = select(raw_data.shape, new_src_slice.raw, None)
                    layout_sel = select(dataset.shape, vslice.raw, None)
                    new_vmap = VDSmap(layout_sel.id, fname, dset_name, vs_sel.id)
                    layout.sources.append(new_vmap)

        head, tail = posixpath.split(name)
        tmp_name = "_tmp_" + tail
        tmp_path = posixpath.join(head, tmp_name)
        dtype = raw_data.dtype
        fillvalue = dataset.fillvalue
        if dtype.metadata and (
            "vlen" in dtype.metadata or "h5py_encoding" in dtype.metadata
        ):
            # Variable length string dtype
            # (https://h5py.readthedocs.io/en/2.10.0/strings.html). Setting the
            # fillvalue in this case doesn't work
            # (https://github.com/h5py/h5py/issues/941).
            if fillvalue not in [0, "", b"", None]:
                raise ValueError(
                    "Non-default fillvalue not supported for variable length strings"
                )
            fillvalue = None
        tmp_dataset = group.create_virtual_dataset(
            tmp_path, layout, fillvalue=fillvalue
        )

        for key, val in dataset.attrs.items():
            tmp_dataset.attrs[key] = val

        if not tmp:
            del group[name]
            group.move(tmp_path, name)


def _is_empty(f: VersionedHDF5File, name: str, version: str) -> bool:
    """Return True if the dataset at the given version is empty, False otherwise.

    Assumes the dataset exists in the given verison.

    Parameters
    ----------
    f : VersionedHDF5File
        File where the dataset resides
    name : str
        Name of the dataset
    version : str
        Version of the dataset to check

    Returns
    -------
    bool
        True if the dataset is empty, False otherwise
    """
    return not f["_version_data/versions"][version][name].is_virtual


def _exists_in_version(f: VersionedHDF5File, name: str, version: str) -> bool:
    """Check if a dataset exists in a given version.

    Parameters
    ----------
    f : VersionedHDF5File
        File where the dataset may reside
    name : str
        Name of the dataset
    version : str
        Version of the dataset to check

    Returns
    -------
    bool
        True if the dataset exists in the version, False otherwise
    """
    return name in f["_version_data/versions"][version]


def _all_extant_are_empty(
    f: VersionedHDF5File, name: str, versions: Iterable[str]
) -> bool:
    """Check if the given versions of a dataset are empty.

    Doesn't assume the dataset exists in any version.

    Parameters
    ----------
    f : VersionedHDF5File
        File where the dataset may reside
    name : str
        Name of the dataset
    version : str
        Version of the dataset to check

    Returns
    -------
    bool
        True if any version of the dataset that can be found is empty,
        False if a version exists which is not.
    """
    for version in versions:
        if _exists_in_version(f, name, version):
            if not _is_empty(f, name, version):
                return False
    return True


def _delete_dataset(f: VersionedHDF5File, name: str, versions_to_delete: Iterable[str]):
    """Delete the given dataset from the versions."""
    version_data = f["_version_data"]
    versions = version_data["versions"]

    if name == "versions":
        return

    versions_to_keep = set(versions) - set(versions_to_delete)

    # If the dataset is empty in the versions to delete, we don't
    # need to recreate the raw data, hash table, or virtual datasets.
    if _all_extant_are_empty(f, name, versions_to_delete):
        return

    raw_data_chunks_map = _recreate_raw_data(f, name, versions_to_delete)

    # If the dataset is not in any versions that are being kept, that
    # data must be deleted.
    if not any([name in versions[version] for version in versions_to_keep]):
        del version_data[name]
        return

    # Recreate the hash table.
    _recreate_hashtable(f, name, raw_data_chunks_map)

    # Recreate every virtual dataset in every kept version.
    _recreate_virtual_dataset(f, name, versions_to_keep, raw_data_chunks_map)


def _walk(g: HLObject, prefix: str = "") -> list[str]:
    """Traverse the object tree, returning all `raw_data` datasets.

    We use this instead of version_data.visit(delete_dataset) because
    visit() has trouble with the groups being deleted from under it.

    Parameters
    ----------
    g : HLObject
        Object containing datasets as descendants
    prefix : str
        Prefix to apply to object names; can be used to filter particular descendants

    Returns
    -------
    list[str]
        List of the names of `raw_data` datasets in g
    """
    datasets = []
    for name in g:
        obj = g[name]
        if isinstance(obj, Group):
            if "raw_data" in obj:
                datasets.append(prefix + name)
            else:
                datasets.extend(_walk(obj, prefix + name + "/"))

    return datasets


def delete_versions(
    f: VersionedHDF5File | File, versions_to_delete: str | Iterable[str]
):
    """Completely delete the given versions from a file

    This function should be used instead of deleting the version group
    directly, as this will not delete the underlying data that is unique to
    the version.
    """
    if isinstance(f, VersionedHDF5File):
        f = f.f

    version_data = f["_version_data"]
    if isinstance(versions_to_delete, str):
        versions_to_delete = [versions_to_delete]

    versions = version_data["versions"]

    if "__first_version__" in versions_to_delete:
        raise ValueError("Cannot delete first version")

    for version in versions_to_delete:
        if version not in versions:
            raise ValueError(f"Version {version!r} does not exist")

    current_version = versions.attrs["current_version"]
    while current_version in versions_to_delete:
        current_version = versions[current_version].attrs["prev_version"]

    for name in _walk(version_data):
        _delete_dataset(f, name, versions_to_delete)

    # find new prev_version which was not deleted
    versions_to_delete_set = set(versions_to_delete)
    for version_name in versions:
        if (
            version_name == "__first_version__"
            or version_name in versions_to_delete_set
        ):
            continue
        prev_version = versions[version_name].attrs["prev_version"]
        while prev_version in versions_to_delete_set:
            prev_version = _get_parent(versions, prev_version)
        versions[version_name].attrs["prev_version"] = prev_version

    # delete the version groups to delete
    for version_name in versions_to_delete:
        del versions[version_name]

    versions.attrs["current_version"] = current_version

    # Collect garbage here to handle intermittent slicing
    # issue; see https://github.com/deshaw/versioned-hdf5/pull/277
    # for a discussion about this.
    gc.collect()


def _get_parent(versions, version_name):
    return versions[version_name].attrs["prev_version"]


# Backwards compatibility
delete_version = delete_versions


def modify_metadata(
    f,
    dataset_name,
    *,
    chunks=None,
    compression=None,
    compression_opts=None,
    dtype=None,
    fillvalue=None,
):
    """
    Modify metadata for a versioned dataset in-place.

    The metadata is modified for all versions containing a dataset.

    `f` should be the h5py file or versioned_hdf5 VersionedHDF5File object.

    `dataset_name` is the name of the dataset in the version group(s).

    Metadata that may be modified are

    - `chunks`: must be compatible with the dataset shape
    - `compression`: see `h5py.Group.create_dataset()`
    - `compression_opts`: see `h5py.Group.create_dataset()`
    - `dtype`: all data in the dataset is cast to the new dtype
    - `fillvalue`: see the note below

    If set to `None` (the default), the given metadata is not modified.

    Note for `fillvalue`, all values equal to the old fillvalue are updated to
    be the new fillvalue, regardless of whether they are explicitly stored or
    represented sparsely in the underlying HDF5 dataset. Also note that
    datasets without an explicitly set fillvalue have a default fillvalue
    equal to the default value of the dtype (e.g., 0. for float dtypes).

    """
    if isinstance(f, VersionedHDF5File):
        f = f.f

    def callback(dataset, version_name):
        _chunks = chunks or dataset.chunks
        _fillvalue = fillvalue or dataset.fillvalue

        if isinstance(dataset, DatasetWrapper):
            dataset = dataset.dataset

        name = dataset.name[len(dataset.parent.name) + 1 :]
        if isinstance(dataset, (InMemoryDataset, InMemoryArrayDataset)):
            new_dataset = InMemoryArrayDataset(
                name, dataset[()], tmp_parent, fillvalue=_fillvalue, chunks=_chunks
            )
            if _fillvalue:
                new_dataset[new_dataset == dataset.fillvalue] = _fillvalue
        elif isinstance(dataset, InMemorySparseDataset):
            new_dataset = InMemorySparseDataset(
                name,
                shape=dataset.shape,
                dtype=dataset.dtype,
                parent=tmp_parent,
                chunks=_chunks,
                fillvalue=_fillvalue,
            )
            new_dataset.data_dict = deepcopy(dataset.data_dict)
            if _fillvalue:
                for a in new_dataset.data_dict.values():
                    a[a == dataset.fillvalue] = _fillvalue
        else:
            raise NotImplementedError(type(dataset))

        if compression:
            new_dataset.compression = compression
        if compression_opts:
            new_dataset.compression_opts = compression_opts

        if dtype:
            return new_dataset.as_dtype(name, dtype, tmp_parent)

        return new_dataset

    newf = tmp_group(f)
    tmp_parent = InMemoryGroup(newf.create_group("__tmp_parent__").id)

    try:
        recreate_dataset(f, dataset_name, newf, callback=callback)

        swap(f, newf)
    finally:
        del newf[newf.name]


def swap(old, new):
    """
    Swap every dataset in old with the corresponding one in new

    Datasets in old that aren't in new are ignored.
    """
    move_names = []

    def _move(name, object):
        if isinstance(object, Dataset):
            if name in new:
                move_names.append(name)

    old.visititems(_move)
    for name in move_names:
        if new[name].is_virtual:
            # We cannot simply move virtual datasets, because they will still
            # point to the old raw_data location. So instead, we have to
            # recreate them, pointing to the new raw_data.
            oldd = old[name]
            newd = new[name]

            def _normalize(path):
                return path if path.endswith("/") else path + "/"

            def _replace_prefix(path, name1, name2):
                """Replace the prefix name1 with name2 in path"""
                name1 = _normalize(name1)
                name2 = _normalize(name2)
                return name2 + path[len(name1) :]

            def _new_vds_layout(d, name1, name2):
                """Recreate a VirtualLayout for d, replacing name1 with name2 in the source dset name"""
                virtual_sources = d.virtual_sources()
                layout = VirtualLayout(d.shape, dtype=d.dtype)
                for vmap in virtual_sources:
                    vspace, fname, dset_name, src_space = vmap
                    assert dset_name.startswith(name1)
                    dset_name = _replace_prefix(dset_name, name1, name2)
                    fname = fname.encode("utf-8")
                    new_vmap = VDSmap(vspace, fname, dset_name, src_space)
                    # h5py 3.3 changed the VirtualLayout code. See
                    # https://github.com/h5py/h5py/pull/1905.
                    if hasattr(layout, "sources"):
                        layout.sources.append(new_vmap)
                    else:
                        layout.dcpl.set_virtual(
                            vspace, fname, dset_name.encode("utf-8"), src_space
                        )
                return layout

            old_layout = _new_vds_layout(oldd, old.name, new.name)
            new_layout = _new_vds_layout(newd, new.name, old.name)
            old_fillvalue = old[name].fillvalue
            new_fillvalue = new[name].fillvalue
            old_attrs = dict(old[name].attrs)
            new_attrs = dict(new[name].attrs)
            del old[name]
            old.create_virtual_dataset(name, new_layout, fillvalue=new_fillvalue)
            for k, v in new_attrs.items():
                if isinstance(v, str) and v.startswith(new.name):
                    v = _replace_prefix(v, new.name, old.name)
                old[name].attrs[k] = v
            del new[name]
            new.create_virtual_dataset(name, old_layout, fillvalue=old_fillvalue)
            for k, v in old_attrs.items():
                if isinstance(v, str) and v.startswith(old.name):
                    v = _replace_prefix(v, old.name, new.name)
                new[name].attrs[k] = v
        else:
            # Invalidate any InMemoryGroups that point to these groups
            delete = []
            for bind in _groups:
                if get_name(bind) and (
                    get_name(bind).startswith(get_name(old.id))
                    or get_name(bind).startswith(get_name(new.id))
                ):
                    delete.append(bind)
            for d in delete:
                del _groups[d]
            old.move(name, posixpath.join(new.name, name + "__tmp"))
            new.move(name, posixpath.join(old.name, name))
            new.move(name + "__tmp", name)



# Pure hdf5 version of TimeInMemoryDataset and TimeInMemoryArraDataset
import os

import h5py
import numpy as np


class TimePureHDF5:
    def setup(self):
        self.file = h5py.File("bench.hdf5", "w")
        self.file.create_dataset(
            "data",
            data=np.arange(10000).reshape((100, 10, 10)),
            chunks=(3, 3, 3),
            maxshape=(None, None, None),
        )

    def teardown(self):
        self.file.close()
        os.remove("bench.hdf5")

    def time_getattr(self):
        dataset = self.file["data"]
        dataset[:, 0, 0:6]

    def time_setattr(self):
        dataset = self.file["data"]
        dataset[:, 0, 0:6] = -1

    def time_resize_bigger(self):
        dataset = self.file["data"]
        dataset.resize((100, 100, 100))

    def time_resize_smaller(self):
        dataset = self.file["data"]
        dataset.resize((10, 10, 10))
